@@ -75,6 +75,10 @@
 #ifdef HAVE_GRP_H
 #include <grp.h>
 #endif /* HAVE_GRP_H */
+#ifdef USE_CAPABILITIES
+#include <sys/capability.h>
+#include <sys/prctl.h>
+#endif /* USE_CAPABILITIES */
 
 #include "xmalloc.h"
 #ifdef HAVE_SETRLIMIT
@@ -663,10 +667,13 @@ int main(int argc,char *argv[])
   char *cvs_cmdline=NULL;
   struct cvsd_addrs *tmp,*nxt;
   struct cvsd_cfg *cfg;
+#ifdef USE_CAPABILITIES
+  cap_t caps;
+#endif /* USE_CAPABILITIES */
 
   /* save program name */
   program_name=argv[0];
-  
+
   /* load the default configuration */
   cfg=cfg_new();
 
@@ -709,7 +716,10 @@ int main(int argc,char *argv[])
     log_log(LOG_ERR,"fatal error saving umask in environment (probably a bug)");
     exit(1);
   }
-  log_log(LOG_DEBUG,"debug: cvsenv[CVSUMASK_IDX]: %s",cfg->cvsenv[CVSUMASK_IDX]);
+
+  /* dump environment */
+  for (i=0;cfg->cvsenv[i]!=NULL;i++)
+    log_log(LOG_DEBUG,"debug: cvsenv[%d]: %s",i,cfg->cvsenv[i]);
 
   /* default address */
   if (cfg->addresses==NULL)
@@ -795,6 +805,24 @@ int main(int argc,char *argv[])
   log_log(LOG_DEBUG,"debug: setgroups() not available");
 #endif /* not HAVE_SETGROUPS */
 
+#ifdef USE_CAPABILITIES
+  /* if there are any capabilities defined, set them to be kept
+     across setuid() calls so we can limit them later on */
+  if (cfg->capabilities!=NULL)
+  {
+    if (prctl(PR_SET_KEEPCAPS,1))
+    {
+      log_log(LOG_ERR,"cannot prctl(PR_SET_KEEPCAPS,1): %s",strerror(errno));
+      exit(1);
+    }
+    log_log(LOG_DEBUG,"debug: prctl(PR_SET_KEEPCAPS,1) done");
+    /* dump the current capabilities */
+    caps=cap_get_proc();
+    log_log(LOG_DEBUG,"debug: current capabilities: %s",cap_to_text(caps,NULL));
+    cap_free(caps);
+  }
+#endif /* USE_CAPABILITIES */
+
   /* change to cvs gid */
   if (cfg->gid!=NOGID)
   {
@@ -817,6 +845,30 @@ int main(int argc,char *argv[])
     log_log(LOG_DEBUG,"debug: setuid(%d) done",cfg->uid);
   }
 
+#ifdef USE_CAPABILITIES
+  /* if there are any capabilities defined, limit them now */
+  if (cfg->capabilities!=NULL)
+  {
+    /* dump the current capabilities */
+    caps=cap_get_proc();
+    log_log(LOG_DEBUG,"debug: current capabilities: %s",cap_to_text(caps,NULL));
+    cap_free(caps);
+    /* limit the capabilities */
+    if (cap_set_proc(cfg->capabilities)!=0)
+    {
+      log_log(LOG_ERR,"cannot cap_set_proc(%s): %s",cap_to_text(cfg->capabilities,NULL),strerror(errno));
+      exit(1);
+    }
+    log_log(LOG_DEBUG,"debug: cap_set_proc(%2) done",cap_to_text(cfg->capabilities,NULL));
+    /* we no longer need this so we should free it */
+    cap_free(cfg->capabilities);
+    /* dump the current capabilities */
+    caps=cap_get_proc();
+    log_log(LOG_DEBUG,"debug: current capabilities: %s",cap_to_text(caps,NULL));
+    cap_free(caps);
+  }
+#endif /* USE_CAPABILITIES */
+
   /* build cvs command line */
   j=strlen(cfg->cvscmd)+1;
   for (i=1;cfg->cvsargs[i]!=NULL;i++)
@@ -857,7 +909,7 @@ int main(int argc,char *argv[])
       /* wait for a new connection */
       acceptconnection(cfg);
       /* handle exiting child processes */
-      reap_deadchildren();    
+      reap_deadchildren();
     }
     else /* not ready for connection */
     {
@@ -869,7 +921,7 @@ int main(int argc,char *argv[])
         /* sleep unblocks on a signal */
         sleep(SLEEPSECS);
         /* handle exiting child processes */
-        reap_deadchildren();    
+        reap_deadchildren();
       }
       /* log that the connections are full */
       log_log(LOG_INFO,"resuming accepting new connections");
